/*
 * Copyright Elasticsearch B.V. and/or licensed to Elasticsearch B.V. under one
 * or more contributor license agreements. Licensed under the Elastic License
 * 2.0; you may not use this file except in compliance with the Elastic License
 * 2.0.
 */

import {
  EuiButton,
  EuiComboBox,
  EuiDescribedFormGroup,
  EuiFormRow,
  EuiHorizontalRule,
  EuiLink,
  EuiSpacer,
  EuiText,
  EuiTitle,
} from '@elastic/eui';
import React, { Component, Fragment } from 'react';

import { i18n } from '@kbn/i18n';
import { FormattedMessage } from '@kbn/i18n-react';
import type { PublicMethodsOf } from '@kbn/utility-types';
import type { DocLinksStart } from 'src/core/public';

import type { SecurityLicense } from '../../../../../../common/licensing';
import type { BuiltinESPrivileges, Role } from '../../../../../../common/model';
import type { IndicesAPIClient } from '../../../indices_api_client';
import { CollapsiblePanel } from '../../collapsible_panel';
import type { RoleValidator } from '../../validate_role';
import { ClusterPrivileges } from './cluster_privileges';
import { IndexPrivileges } from './index_privileges';

interface Props {
  role: Role;
  editable: boolean;
  indicesAPIClient: PublicMethodsOf<IndicesAPIClient>;
  docLinks: DocLinksStart;
  license: SecurityLicense;
  onChange: (role: Role) => void;
  runAsUsers: string[];
  validator: RoleValidator;
  builtinESPrivileges: BuiltinESPrivileges;
  indexPatterns: string[];
}

export class ElasticsearchPrivileges extends Component<Props, {}> {
  public render() {
    return (
      <CollapsiblePanel iconType={'logoElasticsearch'} title={'Elasticsearch'}>
        {this.getForm()}
      </CollapsiblePanel>
    );
  }

  public getForm = () => {
    const {
      role,
      indicesAPIClient,
      docLinks,
      validator,
      onChange,
      editable,
      indexPatterns,
      license,
      builtinESPrivileges,
    } = this.props;

    const indexProps = {
      role,
      indicesAPIClient,
      validator,
      indexPatterns,
      license,
      onChange,
      availableIndexPrivileges: builtinESPrivileges.index,
    };

    return (
      <Fragment>
        <EuiDescribedFormGroup
          title={
            <h3>
              <FormattedMessage
                id="xpack.security.management.editRole.elasticSearchPrivileges.clusterPrivilegesTitle"
                defaultMessage="Cluster privileges"
              />
            </h3>
          }
          description={
            <p>
              <FormattedMessage
                id="xpack.security.management.editRole.elasticSearchPrivileges.manageRoleActionsDescription"
                defaultMessage="Manage the actions this role can perform against your cluster. "
              />
              {this.learnMore(docLinks.links.security.clusterPrivileges)}
            </p>
          }
        >
          <EuiFormRow fullWidth={true} hasEmptyLabelSpace>
            <ClusterPrivileges
              role={this.props.role}
              onChange={this.onClusterPrivilegesChange}
              builtinClusterPrivileges={builtinESPrivileges.cluster}
            />
          </EuiFormRow>
        </EuiDescribedFormGroup>

        <EuiSpacer />

        <EuiDescribedFormGroup
          title={
            <h3>
              <FormattedMessage
                id="xpack.security.management.editRole.elasticSearchPrivileges.runAsPrivilegesTitle"
                defaultMessage="Run As privileges"
              />
            </h3>
          }
          description={
            <p>
              <FormattedMessage
                id="xpack.security.management.editRole.elasticSearchPrivileges.howToBeSubmittedOnBehalfOfOtherUsersDescription"
                defaultMessage="Allow requests to be submitted on the behalf of other users. "
              />
              {this.learnMore(docLinks.links.security.runAsPrivilege)}
            </p>
          }
        >
          <EuiFormRow hasEmptyLabelSpace>
            <EuiComboBox
              placeholder={
                this.props.editable
                  ? i18n.translate(
                      'xpack.security.management.editRole.elasticSearchPrivileges.addUserTitle',
                      { defaultMessage: 'Add a user…' }
                    )
                  : undefined
              }
              options={this.props.runAsUsers.map((username) => ({
                id: username,
                label: username,
                isGroupLabelOption: false,
              }))}
              selectedOptions={this.props.role.elasticsearch.run_as.map((u) => ({ label: u }))}
              onCreateOption={this.onCreateRunAsOption}
              onChange={this.onRunAsUserChange}
              isDisabled={!editable}
            />
          </EuiFormRow>
        </EuiDescribedFormGroup>

        <EuiSpacer />

        <EuiTitle size={'xs'}>
          <h3>
            <FormattedMessage
              id="xpack.security.management.editRole.elasticSearchPrivileges.indexPrivilegesTitle"
              defaultMessage="Index privileges"
            />
          </h3>
        </EuiTitle>
        <EuiSpacer size={'s'} />
        <EuiText size={'s'} color={'subdued'}>
          <p>
            <FormattedMessage
              id="xpack.security.management.editRole.elasticSearchPrivileges.controlAccessToClusterDataDescription"
              defaultMessage="Control access to the data in your cluster. "
            />
            {this.learnMore(docLinks.links.security.indicesPrivileges)}
          </p>
        </EuiText>

        <IndexPrivileges {...indexProps} />

        <EuiHorizontalRule />

        {this.props.editable && (
          <EuiButton iconType={'plusInCircle'} onClick={this.addIndexPrivilege}>
            <FormattedMessage
              id="xpack.security.management.editRole.elasticSearchPrivileges.addIndexPrivilegesButtonLabel"
              defaultMessage="Add index privilege"
            />
          </EuiButton>
        )}
      </Fragment>
    );
  };

  public learnMore = (href: string) => (
    <EuiLink className="editRole__learnMore" href={href} target={'_blank'}>
      <FormattedMessage
        id="xpack.security.management.editRole.elasticSearchPrivileges.learnMoreLinkText"
        defaultMessage="Learn more"
      />
    </EuiLink>
  );

  public addIndexPrivilege = () => {
    const { role } = this.props;

    const newIndices = [
      ...role.elasticsearch.indices,
      {
        names: [],
        privileges: [],
        field_security: {
          grant: ['*'],
        },
      },
    ];

    this.props.onChange({
      ...this.props.role,
      elasticsearch: {
        ...this.props.role.elasticsearch,
        indices: newIndices,
      },
    });
  };

  public onClusterPrivilegesChange = (cluster: string[]) => {
    const role = {
      ...this.props.role,
      elasticsearch: {
        ...this.props.role.elasticsearch,
        cluster,
      },
    };

    this.props.onChange(role);
  };

  public onRunAsUserChange = (users: any) => {
    const role = {
      ...this.props.role,
      elasticsearch: {
        ...this.props.role.elasticsearch,
        run_as: users.map((u: any) => u.label),
      },
    };

    this.props.onChange(role);
  };

  public onCreateRunAsOption = (option: any) => {
    const newRunAsUsers = this.props.role.elasticsearch.run_as.concat(option);

    const role = {
      ...this.props.role,
      elasticsearch: {
        ...this.props.role.elasticsearch,
        run_as: newRunAsUsers,
      },
    };
    this.props.onChange(role);
  };
}
